#    @Author: Justin C Presley
#    @Author-Email: justincpresley@gmail.com
#    @Project: NDN Storage
#    @Source-Code: https://github.com/justincpresley/ndn-python-storage
#    @Pip-Library: https://pypi.org/project/ndn-storage

# API classes
from .disk import DiskStorage
from .leveldb import LevelDBStorage
from .memory import MemoryStorage
from .mongodb import MongoDBStorage
from .sqlite import SqliteStorage
from .storage import Storage


#    @Author: Justin C Presley
#    @Author-Email: justincpresley@gmail.com
#    @Project: NDN Storage
#    @Source-Code: https://github.com/justincpresley/ndn-python-storage
#    @Pip-Library: https://pypi.org/project/ndn-storage

# Change Notice! This file oringinates from another repository.
# Please look at CHANGES.rst for more details.

# Basic Libraries
import asyncio as aio
from contextlib import suppress
from typing import List, Optional
from abc import abstractmethod
# NDN Imports
from ndn.encoding.tlv_var import parse_tl_num
from ndn.encoding import Name, parse_data, NonStrictName
from ndn.name_tree import NameTrie
# Custom Imports
from .storage import Storage

class DiskStorage(Storage):
    class UninitializedError(Exception):
        pass # raised when DiskStorage is used without initializing the storage.
    class ReinitalizedError(Exception):
        pass # raised when trying to initialize and already initialized DiskStorage.

    def __init__(self, write_period:int=10, initialize:bool=True) -> None:
        super().__init__()
        self.write_period:int = write_period
        self.initialized:bool = initialize
        if initialize == True:
            self._initialize_storage()

    def __del__(self) -> None:
        if self.write_period > 0 and self.initialized:
            self.write_back_task.cancel()

    def _start_writing(self) -> None:
        if self.write_period > 0:
            self.write_back_task = aio.create_task(self._periodic_write_back())

    @abstractmethod
    def _initialize_storage(self) -> None:
        return
    @abstractmethod
    def _put(self, key:bytes, data:bytes, expire_time_ms:int=None) -> None:
        return
    @abstractmethod
    def _put_batch(self, keys:List[bytes], values:List[bytes], expire_time_mss:List[Optional[int]]) -> None:
        return
    @abstractmethod
    def _get(self, key:bytes, can_be_prefix:bool=False, must_be_fresh:bool=False) -> bytes:
        return
    @abstractmethod
    def _remove(self, key:bytes) -> bool:
        return

    @staticmethod
    def _get_name_bytes_wo_tl(name:NonStrictName) -> bytes:
        # remove name's TL as key to support efficient prefix search
        name:bytes = Name.to_bytes(name)
        offset:int = 0
        offset += parse_tl_num(name, offset)[1]
        offset += parse_tl_num(name, offset)[1]
        return name[offset:]

    async def _periodic_write_back(self) -> None:
        with suppress(aio.CancelledError):
            while 1:
                self._write_back()
                await aio.sleep(self.write_period)

    def _write_back(self) -> None:
        keys = []
        values = []
        expire_time_mss = []
        for name, (data, expire_time_ms) in self.cache.iteritems(prefix=[], shallow=True):
            keys.append(self._get_name_bytes_wo_tl(name))
            values.append(data)
            expire_time_mss.append(expire_time_ms)
        if len(keys) > 0:
            self._put_batch(keys, values, expire_time_mss)
        self.cache = NameTrie()

    def initialize(self) -> None:
        if self.initialized == True:
            raise self.ReinitalizedError("The storage was initialized more than once.")
        self._initialize_storage()

    def put_packet(self, name:NonStrictName, data:bytes) -> None:
        if self.initialized != True:
            raise self.UninitializedError("The storage is not initialized.")

        _, meta_info, _, _ = parse_data(data)
        expire_time_ms:int = self._time_ms()
        if meta_info.freshness_period:
            expire_time_ms += meta_info.freshness_period

        name = Name.normalize(name)
        if self.write_period > 0:
            self.cache[name] = (data, expire_time_ms)
        else:
            self._put(self._get_name_bytes_wo_tl(name), data, expire_time_ms)

    def get_packet(self, name:NonStrictName, can_be_prefix:bool=False, must_be_fresh:bool=False) -> Optional[bytes]:
        if self.initialized != True:
            raise self.UninitializedError("The storage is not initialized.")

        name = Name.normalize(name)
        # memory lookup
        try:
            if self.write_period > 0:
                if not can_be_prefix:
                    data, expire_time_ms = self.cache[name]
                    if not must_be_fresh or expire_time_ms > self._time_ms():
                        return data
                else:
                    it = self.cache.itervalues(prefix=name, shallow=True)
                    while True:
                        data, expire_time_ms = next(it)
                        if not must_be_fresh or expire_time_ms > self._time_ms():
                            return data
            else:
                raise KeyError
        # storage lookup
        except (KeyError, StopIteration):
            key = self._get_name_bytes_wo_tl(name)
            return self._get(key, can_be_prefix, must_be_fresh)

    def remove_packet(self, name:NonStrictName) -> bool:
        if self.initialized != True:
            raise self.UninitializedError("The storage is not initialized.")

        removed = False
        name = Name.normalize(name)
        try:
            if self.write_period > 0:
                del self.cache[name]
                removed = True
            else:
                raise KeyError
        except KeyError:
            pass
        if self._remove(self._get_name_bytes_wo_tl(name)):
            removed = True
        return removed


#    @Author: Justin C Presley
#    @Author-Email: justincpresley@gmail.com
#    @Project: NDN Storage
#    @Source-Code: https://github.com/justincpresley/ndn-python-storage
#    @Pip-Library: https://pypi.org/project/ndn-storage

# Change Notice! This file oringinates from another repository.
# Please look at CHANGES.rst for more details.

# Basic Libraries
import os
import pickle
import plyvel
from typing import List, Optional
# Custom Imports
from .disk import DiskStorage

class LevelDBStorage(DiskStorage):
    def __init__(self, directory:str, write_period:int=10, initialize:bool=True) -> None:
        self.db_dir = os.path.expanduser(directory)
        if not os.path.exists(self.db_dir):
            try:
                os.makedirs(self.db_dir)
            except PermissionError:
                raise PermissionError(f'Could not create database directory: {self.db_dir}') from None
        super().__init__(write_period, initialize)

    def _initialize_storage(self) -> None:
        self.db = plyvel.DB(self.db_dir, create_if_missing=True)
        self._start_writing()

    def _put(self, key:bytes, value:bytes, expire_time_ms:Optional[int]=None) -> None:
        self.db.put(key, pickle.dumps((value, expire_time_ms)))

    def _put_batch(self, keys:List[bytes], values:List[bytes], expire_time_mss:List[Optional[int]]) -> None:
        with self.db.write_batch() as b:
            for key, value, expire_time_ms in zip(keys, values, expire_time_mss):
                b.put(key, pickle.dumps((value, expire_time_ms)))

    def _get(self, key:bytes, can_be_prefix:bool=False, must_be_fresh:bool=False) -> bytes:
        if not can_be_prefix:
            record = self.db.get(key)
            if record == None:
                return None
            value, expire_time_ms = pickle.loads(record)
            if not must_be_fresh or expire_time_ms != None and expire_time_ms > int(time.time() * 1000):
                return value
            return None
        for _, v_e in self.db.iterator(prefix=key):
            value, expire_time_ms = pickle.loads(v_e)
            if not must_be_fresh or expire_time_ms != None and expire_time_ms > self.time_ms():
                return value
        return None

    def _remove(self, key:bytes) -> bool:
        if self._get(key) != None:
            self.db.delete(key)
            return True
        return False


#    @Author: Justin C Presley
#    @Author-Email: justincpresley@gmail.com
#    @Project: NDN Storage
#    @Source-Code: https://github.com/justincpresley/ndn-python-storage
#    @Pip-Library: https://pypi.org/project/ndn-storage

# Change Notice! This file oringinates from another repository.
# Please look at CHANGES.rst for more details.

# Basic Libraries
import base64
from pymongo import MongoClient, ReplaceOne
from typing import List, Optional
# Custom Imports
from .disk import DiskStorage

class MongoDBStorage(DiskStorage):
    def __init__(self, db:str, collection:str, write_period:int=10, initialize:bool=True) -> None:
        self._db = db
        self._collection = collection
        self._uri = 'mongodb://127.0.0.1:27017/'
        super().__init__(write_period, initialize)

    def _initialize_storage(self) -> None:
        self.client = MongoClient(self._uri)
        self.client.server_info() # will throw an exception if not connected
        self.c_db = self.client[self._db]
        self.c_collection = self.c_db[self._collection]
        self.c_collection.create_index('key', unique=True)
        self._start_writing()

    def _put(self, key:bytes, value:bytes, expire_time_ms:Optional[int]=None) -> None:
        key = base64.b16encode(key).decode()
        replace = {
            'key': key,
            'value': value,
            'expire_time_ms': expire_time_ms,
        }
        self.c_collection.replace_one({'key': key}, replace, upsert=True)

    def _put_batch(self, keys:List[bytes], values:List[bytes], expire_time_mss:List[Optional[int]]) -> None:
        keys = [base64.b16encode(key).decode() for key in keys]
        replaces = []
        for key, value, expire_time_ms in zip(keys, values, expire_time_mss):
            replaces.append(ReplaceOne({'key': key}, {
                'key': key,
                'value': value,
                'expire_time_ms': expire_time_ms,
            }, upsert=True))
        self.c_collection.bulk_write(replaces, ordered=False)

    def _get(self, key:bytes, can_be_prefix:bool=False, must_be_fresh:bool=False) -> Optional[bytes]:
        key = base64.b16encode(key).decode()
        query = dict()
        if not can_be_prefix:
            query.update({'key': key})
        else:
            query.update({'key': {'$regex': '^' + key}})
        if must_be_fresh:
            query.update({'expire_time_ms': {'$gt': self.time_ms()}})
        ret = self.c_collection.find_one(query)
        if ret:
            return ret['value']
        return None

    def _remove(self, key:bytes) -> bool:
        key = base64.b16encode(key).decode()
        return self.c_collection.delete_one({"key": key}).deleted_count > 0


#    @Author: Justin C Presley
#    @Author-Email: justincpresley@gmail.com
#    @Project: NDN Storage
#    @Source-Code: https://github.com/justincpresley/ndn-python-storage
#    @Pip-Library: https://pypi.org/project/ndn-storage

# Basic Libraries
from abc import ABC, abstractmethod
from typing import Optional
from time import time as get_time
# NDN Imports
from ndn.name_tree import NameTrie
from ndn.encoding import NonStrictName

class Storage(ABC):
    def __init__(self) -> None:
        self.cache:NameTrie = NameTrie()
    @staticmethod
    def _time_ms() -> int:
        return int(get_time() * 1000)
    @abstractmethod
    def put_packet(self, name:NonStrictName, packet:bytes) -> None:
        return
    @abstractmethod
    def get_packet(self, name:NonStrictName, can_be_prefix:bool=False, must_be_fresh:bool=False) -> Optional[bytes]:
        return
    @abstractmethod
    def remove_packet(self, name:NonStrictName) -> bool:
        return